"""

Protocol AST declaration.

"""

from enum import Enum
from abc import ABC
from typing import List, Optional


FRAME_ID = b"COZ\x03RE\x01"
MIN_FRAME_SIZE = len(FRAME_ID) + 1 + 2 + 2 + 2


class FrameType(Enum):
    RESET = 1
    RESET_ACK = 2
    FIN = 3
    UNKNOWN_04 = 4
    ENGINE = 7
    ROBOT = 9
    PING = 0x0b


class PacketType(Enum):
    UNKNOWN = -1
    CONNECT = 2
    DISCONNECT = 3
    ACTION = 4
    EVENT = 5
    UNKNOWN_0A = 0x0a
    PING = 0x0b


class Argument(ABC):
    """ Base class for packet arguments. """

    def __init__(self, name: str, description: Optional[str] = None):
        self.name = str(name)
        self.description = str(description) if description else None


class FloatArgument(Argument):
    """ 32-bit floating point number. """

    def __init__(self, name: str, description: Optional[str] = None, default: float = 0.0):
        super().__init__(name, description)
        self.default = float(default)


class DoubleArgument(Argument):
    """ 64-bit floating point number. """

    def __init__(self, name: str, description: Optional[str] = None, default: float = 0.0):
        super().__init__(name, description)
        self.default = float(default)


class BoolArgument(Argument):
    """ 8-bit boolean. """

    def __init__(self, name: str, description: Optional[str] = None, default: bool = False):
        super().__init__(name, description)
        self.default = bool(default)


class UInt8Argument(Argument):
    """ 8-bit unsigned integer. """

    def __init__(self, name: str, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)


class UInt16Argument(Argument):
    """ 16-bit unsigned integer. """

    def __init__(self, name: str, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)


class UInt32Argument(Argument):
    """ 32-bit unsigned integer. """

    def __init__(self, name: str, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)


class Int8Argument(Argument):
    """ 8-bit signed integer. """

    def __init__(self, name: str, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)


class Int16Argument(Argument):
    """ 16-bit signed integer. """

    def __init__(self, name: str, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)


class Int32Argument(Argument):
    """ 32-bit signed integer. """

    def __init__(self, name: str, description: Optional[str] = None, default: int = 0):
        super().__init__(name, description)
        self.default = int(default)


class Packet(ABC):
    """ Base class for packets. """

    def __init__(self, packet_id: PacketType, name: str, description: Optional[str] = None,
                 arguments: Optional[List[Argument]] = None):
        # TODO: Rename to "type".
        self.packet_id = PacketType(packet_id)
        self.name = str(name)
        self.description = str(description) if description else None
        self.arguments = list(arguments) if arguments else []


class Connect(Packet):
    """ Connection acknowledgement packet. """

    def __init__(self):
        super().__init__(PacketType.CONNECT, "connect")


class Disconnect(Packet):
    """ Disconnect packet. """

    def __init__(self):
        super().__init__(PacketType.DISCONNECT, "disconnect")


# TODO: Rename to "Action".
class Command(Packet):
    """ Command packet. """

    def __init__(self, cmd_id: int, name: str, description: Optional[str] = None,
                 arguments: Optional[List[Argument]] = None):
        super().__init__(PacketType.ACTION, name, description, arguments)
        self.id = int(cmd_id)


class Event(Packet):
    """ Event packet. """

    def __init__(self, evt_id: int, name: str, description: Optional[str] = None,
                 arguments: Optional[List[Argument]] = None):
        super().__init__(PacketType.EVENT, name, description, arguments)
        self.id = int(evt_id)


class Ping(Packet):
    """ Ping packet. """

    def __init__(self):
        super().__init__(PacketType.PING, "ping", arguments=[
            DoubleArgument("time_sent_ms"),
            UInt32Argument("counter"),
            UInt32Argument("last"),
            UInt8Argument("unknown"),
        ])


class Unknown0A(Packet):

    def __init__(self):
        super().__init__(PacketType.UNKNOWN_0A, "unknown_0a")


class Protocol(object):
    """ Protocol declaration. """

    def __init__(self, packets: List[Packet]):
        self.packets = list(packets)


PROTOCOL = Protocol(packets=[
    Connect(),
    Disconnect(),
    Ping(),
    Unknown0A(),

    Command(0x0b, "set_head_light", arguments=[
        BoolArgument("enable")
    ]),
    Command(0x32, "drive_wheels", arguments=[
        FloatArgument("lwheel_speed_mmps"),
        FloatArgument("rwheel_speed_mmps"),
        FloatArgument("lwheel_accel_mmps2"),
        FloatArgument("rwheel_accel_mmps2"),
    ]),
    Command(0x33, "turn_in_place", arguments=[
        FloatArgument("wheel_speed_mmps"),
        FloatArgument("wheel_accel_mmps2"),
        Int16Argument("direction"),
    ]),
    Command(0x34, "drive_lift", arguments=[
        FloatArgument("speed"),
    ]),
    Command(0x35, "drive_head", arguments=[
        FloatArgument("speed"),
    ]),
    Command(0x36, "set_lift_height", arguments=[
        FloatArgument("height_mm"),
        FloatArgument("max_speed_rad_per_sec", default=3.0),
        FloatArgument("accel_rad_per_sec2", default=20.0),
        FloatArgument("duration_sec"),
        UInt8Argument("id"),
    ]),
    Command(0x37, "set_head_angle", arguments=[
        FloatArgument("angle_rad"),
        FloatArgument("max_speed_rad_per_sec", default=15.0),
        FloatArgument("accel_rad_per_sec2", default=20.0),
        FloatArgument("duration_sec"),
        UInt8Argument("id"),
    ]),
    Command(0x3b, "stop_all_motors"),

    Command(0xc4, "acknowledge_command", arguments=[
        UInt8Argument("id"),
    ]),

    Command(0xc2, "robot_delocalized"),
    Command(0xc3, "robot_poked"),
])



#!/usr/bin/env python

import sys
import dpkt

import pycozmo


def decode_cozmo_frame(frame_count, ts, buffer):
    frame = pycozmo.Frame.from_bytes(buffer)

    print("{:<12s}first_seq={:04x}, seq={:04x}, ack={:04x}, frame={:6d}, time={:.06f}".format(
        frame.type.name, frame.first_seq, frame.seq, frame.ack, frame_count, ts))

    for pkt in frame.pkts:
        print("\t{}".format(pkt))


def decode_pcap(fspec):
    frame_count = 1
    with open(fspec, "rb") as f:
        first_ts = None
        for ts, frame in dpkt.pcap.Reader(f):
            if first_ts is None:
                first_ts = ts
            rel_ts = ts - first_ts
            eth = dpkt.ethernet.Ethernet(frame)
            if eth.type != dpkt.ethernet.ETH_TYPE_IP:
                # Skip non-IP frames
                continue
            ip = eth.data
            if ip.p != dpkt.ip.IP_PROTO_UDP:
                # Skip non-UDP frames
                continue
            udp = ip.data
            if udp.data[:7] != pycozmo.protocol_declaration.FRAME_ID:
                # Skip non-Cozmo frames
                continue
            decode_cozmo_frame(frame_count, rel_ts, udp.data)
            frame_count += 1
            # if frame_count > 100:
            #     break


def main():
    decode_pcap(sys.argv[1])


if __name__ == '__main__':
    main()




import select
import socket
from datetime import datetime, timedelta
from queue import Queue, Empty
from threading import Thread
from typing import Optional, Tuple, Any

from .frame import Frame
from .protocol_declaration import FrameType
from .protocol_base import Packet, UnknownCommand
from .util import hex_load
from .window import ReceiveWindow, SendWindow
from .protocol_encoder import Connect, Disconnect, Ping


ROBOT_ADDR = ("172.31.1.1", 5551)
PING_INTERVAL = 0.05
RUN_INTERVAL = 0.05


class ReceiveThread(Thread):

    def __init__(self,
                 sock: socket.socket,
                 sender_address: Optional[Tuple[str, int]],
                 timeout: float = 0.5,
                 buffer_size: int = 65536,
                 seq_bits: int = 16,
                 window_size: int = 256) -> None:
        super().__init__(daemon=True, name=__class__.__name__)
        self.sock = sock
        self.sender_address = sender_address
        self.window = ReceiveWindow(seq_bits, size=window_size)
        self.timeout = timeout
        self.buffer_size = buffer_size
        self.stop_flag = False
        self.queue = Queue()

    def stop(self) -> None:
        self.stop_flag = True
        self.join()

    def run(self) -> None:
        while not self.stop_flag:
            ready = select.select([self.sock], [], [], self.timeout)
            if not ready[0]:
                continue

            try:
                raw_frame, address = self.sock.recvfrom(self.buffer_size)
            except InterruptedError:
                continue
            if self.sender_address and self.sender_address != address:
                continue

            frame = Frame.from_bytes(raw_frame)

            self.handle_frame(frame)

    def handle_frame(self, frame: Frame) -> None:
        for pkt in frame.pkts:
            self.handle_pkt(pkt)

    def handle_pkt(self, pkt: Packet) -> None:
        if pkt.is_oob():
            self.deliver(pkt)
            return

        if self.window.is_out_of_order(pkt.seq):
            return

        if self.window.exists(pkt.seq):
            # Duplicate
            return

        self.window.put(pkt.seq, pkt)

        if self.window.is_expected(pkt.seq):
            self.deliver_sequence()

    def deliver_sequence(self) -> None:
        while True:
            pkt = self.window.get()
            if pkt is None:
                break
            self.deliver(pkt)

    def deliver(self, pkt: Packet) -> None:
        self.queue.put(pkt)

    def reset(self):
        self.window.reset()


class SendThread(Thread):

    def __init__(self,
                 sock: socket.socket,
                 receiver_address: Tuple[str, int],
                 timeout: float = 0.5,
                 seq_bits: int = 16,
                 window_size: Optional[int] = 256) -> None:
        super().__init__(daemon=True, name=__class__.__name__)
        self.sock: socket.socket = sock
        self.receiver_address = receiver_address
        self.window = SendWindow(seq_bits, size=window_size)
        self.timeout = timeout
        self.stop_flag = False
        self.queue = Queue()
        self.last_ack = 1
        self.last_send_timestamp = None

    def stop(self) -> None:
        self.stop_flag = True
        self.join()

    def run(self) -> None:
        while not self.stop_flag:
            # if not self.window.is_empty() and \
            #         datetime.now() - self.last_send_timestamp > timedelta(seconds=HELLO_INTERVAL):
            #     pkt = self.window.get_oldest()
            #     pkt.last_ack = self.window.expected_seq
            #     pkt.ack = self.last_ack
            #     frame = pkt.to_bytes()
            #     try:
            #         self.sock.sendto(frame, self.receiver_address)
            #     except InterruptedError:
            #         continue
            #     self.last_send_timestamp = datetime.now()
            #     print("Rsnt {}".format(hex_dump(frame[7:])))
            #     continue

            # if self.window.is_full():
            #     time.sleep(self.timeout)
            #     continue

            try:
                pkt = self.queue.get(timeout=self.timeout)
                self.queue.task_done()
            except Empty:
                continue

            # Construct frame
            if isinstance(pkt, Ping):
                frame = Frame(FrameType.PING, 0, 0, self.last_ack, [pkt])
            else:
                seq = self.window.put(pkt)
                frame = Frame(FrameType.ENGINE, seq, seq, self.last_ack, [pkt])
            raw_frame = frame.to_bytes()

            try:
                self.sock.sendto(raw_frame, self.receiver_address)
            except InterruptedError:
                continue

            self.last_send_timestamp = datetime.now()
            if frame.type != FrameType.PING:
                print("Sent {}".format(pkt))

    def send(self, data: Any) -> None:
        self.queue.put(data)

    def ack(self, seq: int) -> None:
        if self.window.is_out_of_order(seq):
            return

        while self.window.expected_seq <= seq:
            self.window.pop()

    def set_last_ack(self, last_ack: int) -> None:
        self.last_ack = last_ack

    def reset(self) -> None:
        self.window.reset()
        self.last_ack = 1
        self.last_send_timestamp = None


class Client(Thread):

    IDLE = 1
    CONNECTING = 2
    CONNECTED = 3

    def __init__(self, robot_addr: Optional[Tuple[str, int]] = None) -> None:
        super().__init__(daemon=True, name=__class__.__name__)
        self.robot_addr = robot_addr or ROBOT_ADDR
        self.state = self.IDLE
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setblocking(False)
        self.recv_thread = ReceiveThread(self.sock, self.robot_addr)
        self.send_thread = SendThread(self.sock, self.robot_addr)
        self.stop_flag = False
        self.send_last = datetime.now() - timedelta(days=1)

    def start(self) -> None:
        self.recv_thread.start()
        self.send_thread.start()
        super().start()

    def stop(self) -> None:
        self.stop_flag = True
        self.join()
        self.send_thread.stop()
        self.recv_thread.stop()
        self.sock.close()

    def run(self) -> None:
        while not self.stop_flag:
            try:
                pkt = self.recv_thread.queue.get(timeout=RUN_INTERVAL)
                self.recv_thread.queue.task_done()
                self.send_thread.ack(pkt.ack)
                if not pkt.is_oob():
                    self.send_thread.set_last_ack(pkt.seq)
            except Empty:
                pkt = None

            now = datetime.now()
            if self.state == Client.IDLE:
                pass
            elif self.state == Client.CONNECTING:
                if pkt is not None and isinstance(pkt, Connect):
                    print("Connected!")
                    self.state = Client.CONNECTED
            elif self.state == Client.CONNECTED:
                if now - self.send_last > timedelta(seconds=PING_INTERVAL):
                    # print("Sending ping.")
                    self.send_ping()
            else:
                assert False

            if pkt is not None and pkt.PACKET_ID.value not in (5, 0x0b):
                print("Got  {}".format(pkt))

    def connect(self) -> None:
        self.state = self.CONNECTING
        print("Sending hello...")
        self.send_thread.reset()
        self.send_hello()

    def send(self, pkt: Packet):
        self.send_last = datetime.now()
        self.send_thread.send(pkt)

    def send_hello(self) -> None:
        frame = Frame(FrameType.RESET, 1, 1, 0, [])
        raw_frame = frame.to_bytes()

        try:
            self.sock.sendto(raw_frame, self.robot_addr)
        except InterruptedError:
            pass

    def send_disconnect(self) -> None:
        pkt = Disconnect()
        self.send(pkt)

    def send_ping(self) -> None:
        pkt = Ping(0, 1, 0)
        self.send(pkt)

    def send_enable(self) -> None:
        print("Sending enable...")
        pkt = UnknownCommand(0x25)
        self.send(pkt)
        pkt = UnknownCommand(0x4b, b"\xc4\xb69\x00\x00\x00\xa0\xc1")
        self.send(pkt)
        pkt = UnknownCommand(0x9f)
        self.send(pkt)

    def send_init_oled_face(self) -> None:
        pkt = UnknownCommand(0x8f)
        self.send(pkt)
        pkt = UnknownCommand(0x97, hex_load("0d:00:1e:f8:81:f8:83:52:17:f8:81:f8:83:51:1f"))
        self.send(pkt)
        pkt = UnknownCommand(0x8f)
        self.send(pkt)

        pkt = UnknownCommand(0x97, hex_load("0d:00:1e:f8:81:f8:83:53:15:f8:81:f8:83:53:1e"))
        self.send(pkt)
        pkt = UnknownCommand(0x8f)
        self.send(pkt)

        pkt = UnknownCommand(0x97, hex_load("0d:00:1d:f8:81:f8:83:54:15:f8:81:f8:83:54:1d"))
        self.send(pkt)
        pkt = UnknownCommand(0x8f)
        self.send(pkt)

        pkt = UnknownCommand(0x97, hex_load("0d:00:1c:f8:81:f8:83:56:13:f8:81:f8:83:56:1c"))
        self.send(pkt)
        pkt = UnknownCommand(0x8f)
        self.send(pkt)

        pkt = UnknownCommand(0x97, hex_load("14:00:1a:f4:81:40:f4:85:f4:81:83:57:f4:81:40:0f:f8:81:f8:83:57:1c"))
        self.send(pkt)
        pkt = UnknownCommand(0x8f)
        self.send(pkt)

        pkt = UnknownCommand(0x97, hex_load("14:00:19:f4:81:40:f4:85:f4:81:83:58:f4:81:41:0e:f8:81:f8:83:58:1b"))
        self.send(pkt)
        pkt = UnknownCommand(0x8f)
        self.send(pkt)

        pkt = UnknownCommand(0x97, hex_load("14:00:18:f4:81:40:f4:85:f4:81:83:5a:f4:81:41:0c:f8:81:f8:83:5a:1a"))
        self.send(pkt)
        pkt = UnknownCommand(0x8f)
        self.send(pkt)

    def send_led(self) -> None:
        print("Sending LED...")
        pkt = UnknownCommand(0x03, b'\xc8\x90\x86\x88\t\x00\x00\x00\x00\x00\x0a\x99\x86\x88\t\x00\x00\x00\x00\x00\x4c\xa1\x86\x88\t\x00\x00\x00\x00\x00\x00')
        self.send(pkt)
        pkt = UnknownCommand(0x11, b'\x86\x88\x86\x88\t\x00\x00\x00\x00\x00\x8e\xa9\x86\x88\t\x00\x00\x00\x00\x00\x00')
        self.send(pkt)

    def send_big_eyes(self) -> None:
        pkt = UnknownCommand(0x97, hex_load("24:00:16:a0:b6:41:9c:be:40:98:c6:5b:9c:be:9c:a0:b6:40:06:a4:ae:a4:a0:b6:40:9c:be:40:98:c6:5b:9c:be:40:a0:b6:40:16"))
        self.send(pkt)
        pkt = UnknownCommand(0x8f)
        self.send(pkt)

    def send_sleepy_eyes(self) -> None:
        pkt = UnknownCommand(0x97, hex_load("1d:00:17:f4:81:41:f4:85:f4:81:83:5b:f4:81:40:09:f4:81:40:f4:85:f4:81:83:4c:f8:83:4d:f8:81:19"))
        self.send(pkt)
        pkt = UnknownCommand(0x8f)
        self.send(pkt)



"""

Protocol packet encoder classes.

Generated from protocol_declaration.py by protocol_generator.py

Do not modify.

"""

from .protocol_declaration import PacketType
from .protocol_base import Packet
from .protocol_utils import validate_float, validate_bool, validate_integer, get_size, BinaryReader, BinaryWriter

    
class Connect(Packet):

    PACKET_ID = PacketType.CONNECT

    __slots__ = (
    )

    def __init__(self):
        pass

    def __len__(self):
        return 0

    def __repr__(self):
        return "{type}()".format(type=type(self).__name__)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        pass

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        del reader
        return cls(
            )

    
class Disconnect(Packet):

    PACKET_ID = PacketType.DISCONNECT

    __slots__ = (
    )

    def __init__(self):
        pass

    def __len__(self):
        return 0

    def __repr__(self):
        return "{type}()".format(type=type(self).__name__)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        pass

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        del reader
        return cls(
            )

    
class Ping(Packet):

    PACKET_ID = PacketType.PING

    __slots__ = (
        "_time_sent_ms",
        "_counter",
        "_last",
        "_unknown",
    )

    def __init__(self,
                 time_sent_ms=0.0,
                 counter=0,
                 last=0,
                 unknown=0):
        self.time_sent_ms = time_sent_ms
        self.counter = counter
        self.last = last
        self.unknown = unknown

    @property
    def time_sent_ms(self):
        return self._time_sent_ms

    @time_sent_ms.setter
    def time_sent_ms(self, value):
        self._time_sent_ms = validate_float("time_sent_ms", value)

    @property
    def counter(self):
        return self._counter

    @counter.setter
    def counter(self, value):
        self._counter = validate_integer("counter", value, 0, 4294967295)

    @property
    def last(self):
        return self._last

    @last.setter
    def last(self, value):
        self._last = validate_integer("last", value, 0, 4294967295)

    @property
    def unknown(self):
        return self._unknown

    @unknown.setter
    def unknown(self, value):
        self._unknown = validate_integer("unknown", value, 0, 255)

    def __len__(self):
        return \
            get_size('d') + \
            get_size('L') + \
            get_size('L') + \
            get_size('B')

    def __repr__(self):
        return "{type}(" \
               "time_sent_ms={time_sent_ms}, " \
               "counter={counter}, " \
               "last={last}, " \
               "unknown={unknown})".format(
                type=type(self).__name__,
                time_sent_ms=self._time_sent_ms,
                counter=self._counter,
                last=self._last,
                unknown=self._unknown)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        writer.write(self._time_sent_ms, "d")
        writer.write(self._counter, "L")
        writer.write(self._last, "L")
        writer.write(self._unknown, "B")

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        time_sent_ms = reader.read("d")
        counter = reader.read("L")
        last = reader.read("L")
        unknown = reader.read("B")
        return cls(
            time_sent_ms=time_sent_ms,
            counter=counter,
            last=last,
            unknown=unknown)

    
class Unknown0A(Packet):

    PACKET_ID = PacketType.UNKNOWN_0A

    __slots__ = (
    )

    def __init__(self):
        pass

    def __len__(self):
        return 0

    def __repr__(self):
        return "{type}()".format(type=type(self).__name__)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        pass

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        del reader
        return cls(
            )

    
class SetHeadLight(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0x0b

    __slots__ = (
        "_enable",
    )

    def __init__(self,
                 enable=False):
        self.enable = enable

    @property
    def enable(self):
        return self._enable

    @enable.setter
    def enable(self, value):
        self._enable = validate_bool("enable", value)

    def __len__(self):
        return \
            get_size('b')

    def __repr__(self):
        return "{type}(" \
               "enable={enable})".format(
                type=type(self).__name__,
                enable=self._enable)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        writer.write(int(self._enable), "b")

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        enable = bool(reader.read("b"))
        return cls(
            enable=enable)

    
class DriveWheels(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0x32

    __slots__ = (
        "_lwheel_speed_mmps",
        "_rwheel_speed_mmps",
        "_lwheel_accel_mmps2",
        "_rwheel_accel_mmps2",
    )

    def __init__(self,
                 lwheel_speed_mmps=0.0,
                 rwheel_speed_mmps=0.0,
                 lwheel_accel_mmps2=0.0,
                 rwheel_accel_mmps2=0.0):
        self.lwheel_speed_mmps = lwheel_speed_mmps
        self.rwheel_speed_mmps = rwheel_speed_mmps
        self.lwheel_accel_mmps2 = lwheel_accel_mmps2
        self.rwheel_accel_mmps2 = rwheel_accel_mmps2

    @property
    def lwheel_speed_mmps(self):
        return self._lwheel_speed_mmps

    @lwheel_speed_mmps.setter
    def lwheel_speed_mmps(self, value):
        self._lwheel_speed_mmps = validate_float("lwheel_speed_mmps", value)

    @property
    def rwheel_speed_mmps(self):
        return self._rwheel_speed_mmps

    @rwheel_speed_mmps.setter
    def rwheel_speed_mmps(self, value):
        self._rwheel_speed_mmps = validate_float("rwheel_speed_mmps", value)

    @property
    def lwheel_accel_mmps2(self):
        return self._lwheel_accel_mmps2

    @lwheel_accel_mmps2.setter
    def lwheel_accel_mmps2(self, value):
        self._lwheel_accel_mmps2 = validate_float("lwheel_accel_mmps2", value)

    @property
    def rwheel_accel_mmps2(self):
        return self._rwheel_accel_mmps2

    @rwheel_accel_mmps2.setter
    def rwheel_accel_mmps2(self, value):
        self._rwheel_accel_mmps2 = validate_float("rwheel_accel_mmps2", value)

    def __len__(self):
        return \
            get_size('f') + \
            get_size('f') + \
            get_size('f') + \
            get_size('f')

    def __repr__(self):
        return "{type}(" \
               "lwheel_speed_mmps={lwheel_speed_mmps}, " \
               "rwheel_speed_mmps={rwheel_speed_mmps}, " \
               "lwheel_accel_mmps2={lwheel_accel_mmps2}, " \
               "rwheel_accel_mmps2={rwheel_accel_mmps2})".format(
                type=type(self).__name__,
                lwheel_speed_mmps=self._lwheel_speed_mmps,
                rwheel_speed_mmps=self._rwheel_speed_mmps,
                lwheel_accel_mmps2=self._lwheel_accel_mmps2,
                rwheel_accel_mmps2=self._rwheel_accel_mmps2)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        writer.write(self._lwheel_speed_mmps, "f")
        writer.write(self._rwheel_speed_mmps, "f")
        writer.write(self._lwheel_accel_mmps2, "f")
        writer.write(self._rwheel_accel_mmps2, "f")

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        lwheel_speed_mmps = reader.read("f")
        rwheel_speed_mmps = reader.read("f")
        lwheel_accel_mmps2 = reader.read("f")
        rwheel_accel_mmps2 = reader.read("f")
        return cls(
            lwheel_speed_mmps=lwheel_speed_mmps,
            rwheel_speed_mmps=rwheel_speed_mmps,
            lwheel_accel_mmps2=lwheel_accel_mmps2,
            rwheel_accel_mmps2=rwheel_accel_mmps2)

    
class TurnInPlace(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0x33

    __slots__ = (
        "_wheel_speed_mmps",
        "_wheel_accel_mmps2",
        "_direction",
    )

    def __init__(self,
                 wheel_speed_mmps=0.0,
                 wheel_accel_mmps2=0.0,
                 direction=0):
        self.wheel_speed_mmps = wheel_speed_mmps
        self.wheel_accel_mmps2 = wheel_accel_mmps2
        self.direction = direction

    @property
    def wheel_speed_mmps(self):
        return self._wheel_speed_mmps

    @wheel_speed_mmps.setter
    def wheel_speed_mmps(self, value):
        self._wheel_speed_mmps = validate_float("wheel_speed_mmps", value)

    @property
    def wheel_accel_mmps2(self):
        return self._wheel_accel_mmps2

    @wheel_accel_mmps2.setter
    def wheel_accel_mmps2(self, value):
        self._wheel_accel_mmps2 = validate_float("wheel_accel_mmps2", value)

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, value):
        self._direction = validate_integer("direction", value, -32768, 32767)

    def __len__(self):
        return \
            get_size('f') + \
            get_size('f') + \
            get_size('h')

    def __repr__(self):
        return "{type}(" \
               "wheel_speed_mmps={wheel_speed_mmps}, " \
               "wheel_accel_mmps2={wheel_accel_mmps2}, " \
               "direction={direction})".format(
                type=type(self).__name__,
                wheel_speed_mmps=self._wheel_speed_mmps,
                wheel_accel_mmps2=self._wheel_accel_mmps2,
                direction=self._direction)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        writer.write(self._wheel_speed_mmps, "f")
        writer.write(self._wheel_accel_mmps2, "f")
        writer.write(self._direction, "h")

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        wheel_speed_mmps = reader.read("f")
        wheel_accel_mmps2 = reader.read("f")
        direction = reader.read("h")
        return cls(
            wheel_speed_mmps=wheel_speed_mmps,
            wheel_accel_mmps2=wheel_accel_mmps2,
            direction=direction)

    
class DriveLift(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0x34

    __slots__ = (
        "_speed",
    )

    def __init__(self,
                 speed=0.0):
        self.speed = speed

    @property
    def speed(self):
        return self._speed

    @speed.setter
    def speed(self, value):
        self._speed = validate_float("speed", value)

    def __len__(self):
        return \
            get_size('f')

    def __repr__(self):
        return "{type}(" \
               "speed={speed})".format(
                type=type(self).__name__,
                speed=self._speed)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        writer.write(self._speed, "f")

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        speed = reader.read("f")
        return cls(
            speed=speed)

    
class DriveHead(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0x35

    __slots__ = (
        "_speed",
    )

    def __init__(self,
                 speed=0.0):
        self.speed = speed

    @property
    def speed(self):
        return self._speed

    @speed.setter
    def speed(self, value):
        self._speed = validate_float("speed", value)

    def __len__(self):
        return \
            get_size('f')

    def __repr__(self):
        return "{type}(" \
               "speed={speed})".format(
                type=type(self).__name__,
                speed=self._speed)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        writer.write(self._speed, "f")

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        speed = reader.read("f")
        return cls(
            speed=speed)

    
class SetLiftHeight(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0x36

    __slots__ = (
        "_height_mm",
        "_max_speed_rad_per_sec",
        "_accel_rad_per_sec2",
        "_duration_sec",
        "_id",
    )

    def __init__(self,
                 height_mm=0.0,
                 max_speed_rad_per_sec=3.0,
                 accel_rad_per_sec2=20.0,
                 duration_sec=0.0,
                 id=0):
        self.height_mm = height_mm
        self.max_speed_rad_per_sec = max_speed_rad_per_sec
        self.accel_rad_per_sec2 = accel_rad_per_sec2
        self.duration_sec = duration_sec
        self.id = id

    @property
    def height_mm(self):
        return self._height_mm

    @height_mm.setter
    def height_mm(self, value):
        self._height_mm = validate_float("height_mm", value)

    @property
    def max_speed_rad_per_sec(self):
        return self._max_speed_rad_per_sec

    @max_speed_rad_per_sec.setter
    def max_speed_rad_per_sec(self, value):
        self._max_speed_rad_per_sec = validate_float("max_speed_rad_per_sec", value)

    @property
    def accel_rad_per_sec2(self):
        return self._accel_rad_per_sec2

    @accel_rad_per_sec2.setter
    def accel_rad_per_sec2(self, value):
        self._accel_rad_per_sec2 = validate_float("accel_rad_per_sec2", value)

    @property
    def duration_sec(self):
        return self._duration_sec

    @duration_sec.setter
    def duration_sec(self, value):
        self._duration_sec = validate_float("duration_sec", value)

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = validate_integer("id", value, 0, 255)

    def __len__(self):
        return \
            get_size('f') + \
            get_size('f') + \
            get_size('f') + \
            get_size('f') + \
            get_size('B')

    def __repr__(self):
        return "{type}(" \
               "height_mm={height_mm}, " \
               "max_speed_rad_per_sec={max_speed_rad_per_sec}, " \
               "accel_rad_per_sec2={accel_rad_per_sec2}, " \
               "duration_sec={duration_sec}, " \
               "id={id})".format(
                type=type(self).__name__,
                height_mm=self._height_mm,
                max_speed_rad_per_sec=self._max_speed_rad_per_sec,
                accel_rad_per_sec2=self._accel_rad_per_sec2,
                duration_sec=self._duration_sec,
                id=self._id)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        writer.write(self._height_mm, "f")
        writer.write(self._max_speed_rad_per_sec, "f")
        writer.write(self._accel_rad_per_sec2, "f")
        writer.write(self._duration_sec, "f")
        writer.write(self._id, "B")

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        height_mm = reader.read("f")
        max_speed_rad_per_sec = reader.read("f")
        accel_rad_per_sec2 = reader.read("f")
        duration_sec = reader.read("f")
        id = reader.read("B")
        return cls(
            height_mm=height_mm,
            max_speed_rad_per_sec=max_speed_rad_per_sec,
            accel_rad_per_sec2=accel_rad_per_sec2,
            duration_sec=duration_sec,
            id=id)

    
class SetHeadAngle(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0x37

    __slots__ = (
        "_angle_rad",
        "_max_speed_rad_per_sec",
        "_accel_rad_per_sec2",
        "_duration_sec",
        "_id",
    )

    def __init__(self,
                 angle_rad=0.0,
                 max_speed_rad_per_sec=15.0,
                 accel_rad_per_sec2=20.0,
                 duration_sec=0.0,
                 id=0):
        self.angle_rad = angle_rad
        self.max_speed_rad_per_sec = max_speed_rad_per_sec
        self.accel_rad_per_sec2 = accel_rad_per_sec2
        self.duration_sec = duration_sec
        self.id = id

    @property
    def angle_rad(self):
        return self._angle_rad

    @angle_rad.setter
    def angle_rad(self, value):
        self._angle_rad = validate_float("angle_rad", value)

    @property
    def max_speed_rad_per_sec(self):
        return self._max_speed_rad_per_sec

    @max_speed_rad_per_sec.setter
    def max_speed_rad_per_sec(self, value):
        self._max_speed_rad_per_sec = validate_float("max_speed_rad_per_sec", value)

    @property
    def accel_rad_per_sec2(self):
        return self._accel_rad_per_sec2

    @accel_rad_per_sec2.setter
    def accel_rad_per_sec2(self, value):
        self._accel_rad_per_sec2 = validate_float("accel_rad_per_sec2", value)

    @property
    def duration_sec(self):
        return self._duration_sec

    @duration_sec.setter
    def duration_sec(self, value):
        self._duration_sec = validate_float("duration_sec", value)

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = validate_integer("id", value, 0, 255)

    def __len__(self):
        return \
            get_size('f') + \
            get_size('f') + \
            get_size('f') + \
            get_size('f') + \
            get_size('B')

    def __repr__(self):
        return "{type}(" \
               "angle_rad={angle_rad}, " \
               "max_speed_rad_per_sec={max_speed_rad_per_sec}, " \
               "accel_rad_per_sec2={accel_rad_per_sec2}, " \
               "duration_sec={duration_sec}, " \
               "id={id})".format(
                type=type(self).__name__,
                angle_rad=self._angle_rad,
                max_speed_rad_per_sec=self._max_speed_rad_per_sec,
                accel_rad_per_sec2=self._accel_rad_per_sec2,
                duration_sec=self._duration_sec,
                id=self._id)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        writer.write(self._angle_rad, "f")
        writer.write(self._max_speed_rad_per_sec, "f")
        writer.write(self._accel_rad_per_sec2, "f")
        writer.write(self._duration_sec, "f")
        writer.write(self._id, "B")

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        angle_rad = reader.read("f")
        max_speed_rad_per_sec = reader.read("f")
        accel_rad_per_sec2 = reader.read("f")
        duration_sec = reader.read("f")
        id = reader.read("B")
        return cls(
            angle_rad=angle_rad,
            max_speed_rad_per_sec=max_speed_rad_per_sec,
            accel_rad_per_sec2=accel_rad_per_sec2,
            duration_sec=duration_sec,
            id=id)

    
class StopAllMotors(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0x3b

    __slots__ = (
    )

    def __init__(self):
        pass

    def __len__(self):
        return 0

    def __repr__(self):
        return "{type}()".format(type=type(self).__name__)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        pass

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        del reader
        return cls(
            )

    
class AcknowledgeCommand(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0xc4

    __slots__ = (
        "_id",
    )

    def __init__(self,
                 id=0):
        self.id = id

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = validate_integer("id", value, 0, 255)

    def __len__(self):
        return \
            get_size('B')

    def __repr__(self):
        return "{type}(" \
               "id={id})".format(
                type=type(self).__name__,
                id=self._id)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        writer.write(self._id, "B")

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        id = reader.read("B")
        return cls(
            id=id)

    
class RobotDelocalized(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0xc2

    __slots__ = (
    )

    def __init__(self):
        pass

    def __len__(self):
        return 0

    def __repr__(self):
        return "{type}()".format(type=type(self).__name__)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        pass

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        del reader
        return cls(
            )

    
class RobotPoked(Packet):

    PACKET_ID = PacketType.ACTION
    ID = 0xc3

    __slots__ = (
    )

    def __init__(self):
        pass

    def __len__(self):
        return 0

    def __repr__(self):
        return "{type}()".format(type=type(self).__name__)

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
        pass

    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
        del reader
        return cls(
            )


ACTION_BY_ID = {
    0x0b: SetHeadLight,  # 11
    0x32: DriveWheels,  # 50
    0x33: TurnInPlace,  # 51
    0x34: DriveLift,  # 52
    0x35: DriveHead,  # 53
    0x36: SetLiftHeight,  # 54
    0x37: SetHeadAngle,  # 55
    0x3b: StopAllMotors,  # 59
    0xc2: RobotDelocalized,  # 194
    0xc3: RobotPoked,  # 195
    0xc4: AcknowledgeCommand,  # 196
}




import math
from typing import Optional, Any


class BaseWindow(object):

    def __init__(self, seq_bits: int, size: Optional[int] = None) -> None:
        if size is None:
            self.size = int(math.pow(2, seq_bits - 1))
        elif size < 0 or size > int(math.pow(2, seq_bits - 1)):
            raise ValueError("Invalid window size.")
        else:
            self.size = size
        self.expected_seq = 1
        self.last_seq = 0
        self.max_seq = int(math.pow(2, seq_bits))

    def is_valid_seq(self, seq: int) -> bool:
        res = 0 <= seq < self.max_seq
        return res

    def reset(self) -> None:
        self.expected_seq = 1
        self.last_seq = 0


class ReceiveWindow(BaseWindow):

    def __init__(self, seq_bits: int, size: Optional[int] = None) -> None:
        super().__init__(seq_bits, size)
        self.window = [None for _ in range(self.size)]

    def is_out_of_order(self, seq: int) -> bool:
        if self.expected_seq > self.last_seq:
            res = self.expected_seq > seq > self.last_seq
        else:
            res = seq < self.expected_seq or seq > self.last_seq
        return res

    def exists(self, seq: int) -> bool:
        res = self.window[seq % self.size] is not None
        return res

    def put(self, seq: int, data: Any) -> None:
        self.window[seq % self.size] = data

    def is_expected(self, seq: int) -> bool:
        res = seq == self.expected_seq
        return res

    def get(self) -> Any:
        seq = self.expected_seq
        data = self.window[seq % self.size]
        if data is not None:
            self.window[seq % self.size] = None
            self.expected_seq = (seq + 1) % self.max_seq
            self.last_seq = (self.expected_seq + self.size - 1) % self.max_seq
        return data

    def reset(self) -> None:
        super().reset()
        for i in range(self.size):
            self.window[i] = None


class SendWindowSlot(object):

    def __init__(self):
        self.seq: Optional[int] = None
        self.data: Any = None

    def set(self, seq: int, data: Any):
        self.seq = seq
        self.data = data

    def reset(self):
        self.seq = None
        self.data = None


class SendWindow(BaseWindow):

    def __init__(self, seq_bits: int, size: Optional[int] = None) -> None:
        super().__init__(seq_bits, size)
        self.next_seq = 1
        self.window = [SendWindowSlot() for _ in range(self.size)]

    def is_out_of_order(self, seq: int) -> bool:
        if self.is_empty():
            res = True
        elif self.expected_seq > self.next_seq:
            res = self.expected_seq > seq >= self.next_seq
        else:
            res = seq < self.expected_seq or seq >= self.next_seq
        return res

    def is_empty(self) -> bool:
        res = self.expected_seq == self.next_seq
        return res

    def is_full(self) -> bool:
        if self.expected_seq > self.next_seq:
            res = (self.next_seq + self.max_seq - self.expected_seq) >= self.size
        else:
            res = (self.next_seq - self.expected_seq) >= self.size
        return res

    def put(self, data: Any) -> int:
        seq = self.next_seq
        self.next_seq = (self.next_seq + 1) % self.max_seq
        self.window[seq % self.size].set(seq, data)
        return seq

    def pop(self) -> None:
        self.window[self.expected_seq % self.size].reset()
        self.expected_seq = (self.expected_seq + 1) % self.max_seq
        self.last_seq = (self.last_seq + 1) % self.max_seq

    def get_oldest(self) -> Any:
        res = self.window[self.expected_seq % self.size].data
        return res

    def reset(self):
        super().reset()
        self.next_seq = 1
        for slot in self.window:
            slot.reset()




import unittest

from pycozmo.window import BaseWindow, ReceiveWindow, SendWindow


class TestBaseWindowCreate(unittest.TestCase):

    def test_create(self):
        w = BaseWindow(4)
        self.assertEqual(w.size, 8)
        self.assertEqual(w.expected_seq, 1)
        self.assertEqual(w.last_seq, 0)
        self.assertEqual(w.max_seq, 16)

    def test_create_limted(self):
        w = ReceiveWindow(4, size=5)
        self.assertEqual(w.size, 5)
        self.assertEqual(w.expected_seq, 1)
        self.assertEqual(w.last_seq, 0)
        self.assertEqual(w.max_seq, 16)

    def test_create_invalid(self):
        with self.assertRaises(ValueError):
            ReceiveWindow(3, size=10)


class TestBaseWindow(unittest.TestCase):

    def setUp(self):
        self.w = BaseWindow(4)

    def test_is_valid_seq_negative(self):
        self.assertFalse(self.w.is_valid_seq(-1))

    def test_is_valid_seq_too_large(self):
        self.assertFalse(self.w.is_valid_seq(16))

    def test_is_valid_seq(self):
        self.assertTrue(self.w.is_valid_seq(0))


class TestReceiveWindow(unittest.TestCase):

    def setUp(self):
        self.w = ReceiveWindow(4)

    def test_is_out_of_order(self):
        self.assertFalse(self.w.is_out_of_order(0))
        self.assertFalse(self.w.is_out_of_order(1))
        self.assertFalse(self.w.is_out_of_order(10))

    def test_exists(self):
        self.assertFalse(self.w.exists(0))
        self.assertFalse(self.w.exists(1))
        self.w.put(1, "1")
        self.assertFalse(self.w.exists(0))
        self.assertTrue(self.w.exists(1))

    def test_is_expected(self):
        self.assertTrue(self.w.is_expected(1))
        self.assertFalse(self.w.is_expected(2))

    def test_get(self):
        self.assertIsNone(self.w.get())
        self.w.put(2, "2")
        self.w.put(1, "1")
        self.assertEqual(self.w.get(), "1")
        self.assertEqual(self.w.get(), "2")
        self.assertIsNone(self.w.get())


class TestTransmitWindow(unittest.TestCase):

    def setUp(self):
        self.w = SendWindow(4)

    def test_is_out_of_order_empty(self):
        self.assertTrue(self.w.is_out_of_order(0))
        self.assertTrue(self.w.is_out_of_order(1))
        self.assertTrue(self.w.is_out_of_order(15))
        self.assertTrue(self.w.is_out_of_order(16))

    def test_is_out_of_order_one(self):
        self.w.put("1")
        self.assertTrue(self.w.is_out_of_order(0))
        self.assertFalse(self.w.is_out_of_order(1))
        self.assertTrue(self.w.is_out_of_order(2))
        self.assertTrue(self.w.is_out_of_order(16))

    def test_is_empty(self):
        self.assertTrue(self.w.is_empty())
        self.w.put("0")
        self.assertFalse(self.w.is_empty())
        self.w.pop()
        self.assertTrue(self.w.is_empty())

    def test_is_full(self):
        for i in range(8):
            self.assertFalse(self.w.is_full())
            self.w.put(str(i))
        self.assertTrue(self.w.is_full())
        for _ in range(8):
            self.w.pop()
            self.assertFalse(self.w.is_full())




from abc import ABC, abstractmethod

from .protocol_declaration import PacketType
from .protocol_utils import BinaryReader, BinaryWriter
from .util import hex_dump


class Packet(ABC):

    # TODO: Rename to "PACKET_TYPE".
    PACKET_ID = None
    # TODO: Remove.
    seq = 0
    ack = 0

    @abstractmethod
    def __len__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError

    @abstractmethod
    def to_bytes(self) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def to_writer(self, writer: BinaryWriter) -> None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_bytes(cls, buffer: bytes):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_reader(cls, reader: BinaryReader):
        raise NotImplementedError

    def is_oob(self) -> bool:
        res = self.PACKET_ID.value >= PacketType.EVENT.value
        return res


class UnknownPacket(Packet):

    __slots__ = (
        "_PACKET_ID",
        "_data",
    )

    def __init__(self, packet_id: PacketType, data: bytes):
        self.PACKET_ID = packet_id
        self.data = data

    @property
    def PACKET_ID(self):
        return self._PACKET_ID

    @PACKET_ID.setter
    def PACKET_ID(self, value):
        self._PACKET_ID = PacketType(value)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = bytes(value)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return "{type}({id:02x}, {data})".format(
            id=self._PACKET_ID.value, type=type(self).__name__, data=hex_dump(data=self._data))

    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()

    def to_writer(self, writer):
        writer.write_farray(self._data, "B", len(self._data))

    @classmethod
    def from_bytes(cls, buffer):
        # The size is not known.
        raise NotImplementedError

    @classmethod
    def from_reader(cls, reader):
        # The size is not known.
        raise NotImplementedError


class UnknownCommand(UnknownPacket):

    __slots__ = (
        "_PACKET_ID",
        "_ID",
        "_data",
    )

    def __init__(self, cmd_id: int, data: bytes = b""):
        super().__init__(PacketType.ACTION, data)
        self.ID = cmd_id

    @property
    def ID(self):
        return self._ID

    @ID.setter
    def ID(self, value):
        self._ID = int(value)

    @classmethod
    def from_bytes(cls, buffer):
        # The size is not known.
        raise NotImplementedError

    @classmethod
    def from_reader(cls, reader):
        # The size is not known.
        raise NotImplementedError

    def __repr__(self):
        return "{type}({id:02x}, {data})".format(
            id=self._ID, type=type(self).__name__, data=hex_dump(data=self._data))


class UnknownEvent(UnknownPacket):

    __slots__ = (
        "_PACKET_ID",
        "_ID",
        "_data",
    )

    def __init__(self, cmd_id: int, data: bytes = b""):
        super().__init__(PacketType.EVENT, data)
        self.ID = cmd_id

    @property
    def ID(self):
        return self._ID

    @ID.setter
    def ID(self, value):
        self._ID = int(value)

    @classmethod
    def from_bytes(cls, buffer):
        # The size is not known.
        raise NotImplementedError

    @classmethod
    def from_reader(cls, reader):
        # The size is not known.
        raise NotImplementedError

    def __repr__(self):
        return "{type}({id:02x}, {data})".format(
            id=self._ID, type=type(self).__name__, data=hex_dump(data=self._data))




def hex_dump(data: bytes) -> str:
    res = ":".join("{:02x}".format(b) for b in data)
    return res


def hex_load(data: str) -> bytes:
    res = bytearray.fromhex(data.replace(":", ""))
    return res


def frange(start, stop, step):
    x = start
    while x < stop:
        yield x
        x += step



"""

Protocol packet encoder code generator.

"""

import os

from . import protocol_declaration


def to_pascal_case(name):
    components = name.split('_')
    return ''.join(x.title() for x in components)


class ProtocolGenerator(object):

    def __init__(self, f):
        self.f = f

    def generate_action_map(self):
        action_map = {}
        for packet in protocol_declaration.PROTOCOL.packets:
            if not isinstance(packet, protocol_declaration.Command):
                continue
            action_map[packet.id] = to_pascal_case(packet.name)

        self.f.write('\n\nACTION_BY_ID = {\n')
        for k, v in sorted(action_map.items()):
            self.f.write('    0x{id:02x}: {name},  # {id}\n'.format(id=k, name=v))
        self.f.write('}\n')

    def generate_packet_slots(self, packet):
        for argument in packet.arguments:
            self.f.write('        "_{name}",\n'.format(name=argument.name))

    def generate_argument_methods(self, packet):
        for argument in packet.arguments:
            self.f.write(r"""
    @property
    def {name}(self):
        return self._{name}

    @{name}.setter
    def {name}(self, value):
        self._{name} = """.format(name=argument.name))
            if isinstance(argument, protocol_declaration.FloatArgument):
                self.f.write('validate_float("{name}", value)\n'.format(name=argument.name))
            elif isinstance(argument, protocol_declaration.DoubleArgument):
                self.f.write('validate_float("{name}", value)\n'.format(name=argument.name))
            elif isinstance(argument, protocol_declaration.BoolArgument):
                self.f.write('validate_bool("{name}", value)\n'.format(name=argument.name))
            elif isinstance(argument, protocol_declaration.UInt8Argument):
                self.f.write('validate_integer("{name}", value, 0, 255)\n'.format(name=argument.name))
            elif isinstance(argument, protocol_declaration.UInt32Argument):
                self.f.write('validate_integer("{name}", value, 0, 4294967295)\n'.format(name=argument.name))
            elif isinstance(argument, protocol_declaration.Int16Argument):
                self.f.write('validate_integer("{name}", value, -32768, 32767)\n'.format(name=argument.name))
            else:
                raise NotImplementedError("Unexpected argument type '{}' for '{}'".format(
                    type(argument), argument.name))

    def generate_len_method(self, packet):
        self.f.write("\n    def __len__(self):\n")
        if not packet.arguments:
            self.f.write("        return 0\n")
        else:
            self.f.write("        return \\\n")
            statements = []
            for argument in packet.arguments:
                if isinstance(argument, protocol_declaration.FloatArgument):
                    statements.append("get_size('f')")
                elif isinstance(argument, protocol_declaration.DoubleArgument):
                    statements.append("get_size('d')")
                elif isinstance(argument, protocol_declaration.BoolArgument):
                    statements.append("get_size('b')")
                elif isinstance(argument, protocol_declaration.UInt8Argument):
                    statements.append("get_size('B')")
                elif isinstance(argument, protocol_declaration.UInt32Argument):
                    statements.append("get_size('L')")
                elif isinstance(argument, protocol_declaration.Int16Argument):
                    statements.append("get_size('h')")
                else:
                    raise NotImplementedError("Unexpected argument type '{}' for '{}'".format(
                        type(argument), argument.name))
            self.f.write("            ")
            self.f.write(" + \\\n            ".join(statements))
            self.f.write("\n")

    def generate_repr_method(self, packet):
        self.f.write("\n    def __repr__(self):\n")
        if packet.arguments:
            argument_strs = []
            arguments = ["type=type(self).__name__"]
            if packet.arguments:
                for argument in packet.arguments:
                    argument_strs.append('"{name}={{{name}}}'.format(name=argument.name))
                    arguments.append("{name}=self._{name}".format(name=argument.name))
            self.f.write('        return "{type}(" \\\n               ')
            self.f.write('{argument_strs})".format(\n                {arguments})\n'.format(
                argument_strs=', " \\\n               '.join(argument_strs),
                arguments=",\n                ".join(arguments)))
        else:
            self.f.write('        return "{type}()".format(type=type(self).__name__)\n')

    def generate_argument_defaults(self, packet):
        for argument in packet.arguments:
            self.f.write(",\n                 {name}={default}".format(name=argument.name, default=argument.default))

    def generate_arugment_assignments(self, packet):
        if packet.arguments:
            for argument in packet.arguments:
                self.f.write("        self.{name} = {name}\n".format(name=argument.name))
        else:
            self.f.write("        pass\n")

    def generate_packet_encoding(self, packet):
        self.f.write(r"""
    def to_bytes(self):
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()
        
    def to_writer(self, writer):
""")
        if packet.arguments:
            for argument in packet.arguments:
                if isinstance(argument, protocol_declaration.FloatArgument):
                    self.f.write('        writer.write(self._{name}, "f")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.DoubleArgument):
                    self.f.write('        writer.write(self._{name}, "d")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.BoolArgument):
                    self.f.write('        writer.write(int(self._{name}), "b")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt8Argument):
                    self.f.write('        writer.write(self._{name}, "B")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt32Argument):
                    self.f.write('        writer.write(self._{name}, "L")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int16Argument):
                    self.f.write('        writer.write(self._{name}, "h")\n'.format(name=argument.name))
                else:
                    raise NotImplementedError("Unexpected argument type '{}' for '{}'".format(
                        type(argument), argument.name))
        else:
            self.f.write("        pass\n")

    def generate_packet_decoding(self, packet):
        self.f.write(r"""
    @classmethod
    def from_bytes(cls, buffer):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj
        
    @classmethod
    def from_reader(cls, reader):
""")
        if packet.arguments:
            for argument in packet.arguments:
                if isinstance(argument, protocol_declaration.FloatArgument):
                    self.f.write('        {name} = reader.read("f")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.DoubleArgument):
                    self.f.write('        {name} = reader.read("d")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.BoolArgument):
                    self.f.write('        {name} = bool(reader.read("b"))\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt8Argument):
                    self.f.write('        {name} = reader.read("B")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.UInt32Argument):
                    self.f.write('        {name} = reader.read("L")\n'.format(name=argument.name))
                elif isinstance(argument, protocol_declaration.Int16Argument):
                    self.f.write('        {name} = reader.read("h")\n'.format(name=argument.name))
                else:
                    raise NotImplementedError("Unexpected argument type '{}' for '{}'".format(
                        type(argument), argument.name))
        else:
            self.f.write("        del reader\n")
        self.f.write("        return cls(\n")
        arguments = []
        for argument in packet.arguments:
            arguments.append("{name}={name}".format(name=argument.name))
        self.f.write("            ")
        self.f.write(",\n            ".join(arguments))
        self.f.write(")\n")

    def generate_packet(self, packet):
        self.f.write(r"""
    
class {name}(Packet):

    PACKET_ID = {packet_id}
""".format(name=to_pascal_case(packet.name), packet_id=packet.packet_id))
        if isinstance(packet, protocol_declaration.Command) or isinstance(packet, protocol_declaration.Event):
            self.f.write("    ID = 0x{:02x}\n".format(packet.id))
        self.f.write("\n    __slots__ = (\n")
        self.generate_packet_slots(packet)
        self.f.write("    )\n\n    def __init__(self")
        self.generate_argument_defaults(packet)
        self.f.write("):\n")
        self.generate_arugment_assignments(packet)
        self.generate_argument_methods(packet)
        self.generate_len_method(packet)
        self.generate_repr_method(packet)
        self.generate_packet_encoding(packet)
        self.generate_packet_decoding(packet)

    def generate(self):
        header = r'''"""

Protocol packet encoder classes.

Generated from {declaration} by {generator}

Do not modify.

"""

from .protocol_declaration import PacketType
from .protocol_base import Packet
from .protocol_utils import validate_float, validate_bool, validate_integer, get_size, BinaryReader, BinaryWriter
'''.format(declaration=os.path.basename(protocol_declaration.__file__), generator=os.path.basename(__file__))
        self.f.write(header)

        for packet in protocol_declaration.PROTOCOL.packets:
            self.generate_packet(packet)

        self.generate_action_map()




import struct


_struct_cache = dict()


def _get_struct(fmt, length):
    key = (fmt, length)
    if key in _struct_cache:
        return _struct_cache[key]
    else:
        reader = struct.Struct('<{0}{1}'.format(length, fmt))
        _struct_cache[key] = reader
        return reader


def validate_float(name, value):
    try:
        value = float(value)
    except ValueError:
        raise ValueError("{name} must be float. Got {type}.".format(name=name, type=type(value).__name__))
    return value


def validate_bool(name, value):
    try:
        value = bool(value)
    except ValueError:
        raise ValueError("{name} must be bool. Got {type}.".format(name=name, type=type(value).__name__))
    return value


def validate_integer(name, value, minimum, maximum):
    try:
        value = int(value)
    except ValueError:
        raise ValueError("{name} must be an integer. Got a {type}.".format(name=name, type=type(value).__name__))
    if value < minimum or value > maximum:
        raise ValueError('{name} must be between {minimum} and {maximum}. Got {value}.'.format(
            name=name, minimum=minimum, maximum=maximum, value=value))
    return value


def get_size(fmt):
    """ Figures out the size of a value with the given format. """
    return _get_struct(fmt, 1).size


class BinaryReader(object):
    """ Used to read in a stream of binary data, keeping track of the current position. """

    def __init__(self, buffer: bytes, offset: int = 0):
        self._buffer = buffer
        self._index = offset

    @property
    def buffer(self):
        return self._buffer

    def __len__(self):
        return len(self._buffer)

    def seek_set(self, offset: int):
        if offset < 0 or offset > len(self._buffer):
            ValueError("Invalid offset.")
        self._index = offset

    def seek_cur(self, offset: int):
        offset += self._index
        if offset < 0 or offset > len(self._buffer):
            ValueError("Invalid offset.")
        self._index = offset

    def tell(self):
        """ Returns the current stream position as an offset within the buffer. """
        return self._index

    def read(self, fmt):
        """ Reads in a single value of the given format. """
        return self.read_farray(fmt, 1)[0]

    def read_farray(self, fmt, length):
        """ Reads in a fixed-length array of the given format and length. """
        reader = _get_struct(fmt, length)
        if self._index + reader.size > len(self._buffer):
            raise IndexError('Buffer not large enough to read serialized message. Received {0} bytes.'.format(
                len(self._buffer)))
        result = reader.unpack_from(self._buffer, self._index)
        self._index += reader.size
        return result

    def read_varray(self, data_format, length_format):
        """ Reads in a variable-length array with the given length format and data format. """
        length = self.read(length_format)
        return self.read_farray(data_format, length)

    def read_string(self, length_format):
        """ Reads in a variable-length string with the given length format. """
        length = self.read(length_format)
        bs = self.read_farray('s', length)[0]
        return bs.decode('utf_8')

    def read_string_farray(self, string_length_format, array_length):
        """ Reads in a fixed-length array of variable-length strings with the given length format. """
        return [self.read_string(string_length_format) for _ in range(array_length)]

    def read_string_varray(self, string_length_format, array_length_format):
        """ Reads in a variable-length array of variable-length strings with the given length format. """
        array_length = self.read(array_length_format)
        return [self.read_string(string_length_format) for _ in range(array_length)]

    def read_object(self, from__reader_method):
        """ Reads in an object according to the given method. """
        return from__reader_method(self)

    def read_object_farray(self, from__reader_method, length):
        """ Reads in a fixed-length object sequence according to the given method. """
        return [from__reader_method(self) for _ in range(length)]

    def read_object_varray(self, from__reader_method, length_format):
        """ Reads in a variable-length object sequence according to the given method. """
        length = self.read(length_format)
        return [from__reader_method(self) for _ in range(length)]


class BinaryWriter(object):
    """ Used to write out a stream of binary data. """

    def __init__(self):
        self._buffer = []

    def clear(self):
        del self._buffer[:]

    def dumps(self) -> bytes:
        return b"".join(self._buffer)

    def write_bytes(self, value: bytes):
        """ Writes out a byte sequence. """
        self._buffer.append(value)

    def write(self, value, fmt):
        """ Writes out a single value of the given format. """
        self.write_farray((value,), fmt, 1)

    def write_farray(self, value, fmt, length):
        """ Writes out a fixed-length array of the given format and length. """
        writer = _get_struct(fmt, length)
        self._buffer.append(writer.pack(*value))

    def write_varray(self, value, data_format, length_format):
        """ Writes out a variable-length array with the given length format and data format. """
        self.write(len(value), length_format)
        self.write_farray(value, data_format, len(value))

    def write_string(self, value, length_format):
        """ Writes out a variable-length string with the given length format. """
        bs = value.encode('utf_8')
        self.write(len(bs), length_format)
        self.write_farray((bs,), 's', len(bs))

    def write_string_farray(self, value, string_length_format, array_length):
        """ Writes out a fixed-length array of variable-length strings with the given length format. """
        if len(value) != array_length:
            raise ValueError('The given fixed-length sequence has the wrong length.')
        for element in value:
            self.write_string(element, string_length_format)

    def write_string_varray(self, value, string_length_format, array_length_format):
        """ Writes out a variable-length array of variable-length strings with the given length format. """
        self.write(len(value), array_length_format)
        for element in value:
            self.write_string(element, string_length_format)

    def write_object(self, value):
        """ Writes out an object that supports a to_writer() method. """
        value.to_writer(self)

    def write_object_farray(self, value, length):
        """ Writes out a fixed-length object sequence that supports a to_writer() method. """
        if len(value) != length:
            raise ValueError('The given fixed-length sequence has the wrong length.')
        for element in value:
            element.to_writer(self)

    def write_object_varray(self, value, length_format):
        """ Writes out a variable-length object sequence that supports a to_writer() method. """
        self.write(len(value), length_format)
        for element in value:
            element.to_writer(self)




import unittest

from pycozmo.frame import Frame


class TestFrame(unittest.TestCase):

    def test_from_bytes_multi(self):
        f = Frame.from_bytes(
            b'COZ\x03RE\x01\x07\x9d\n\xa0\n\x8f\x00\x04\x01\x00\x8f\x04\x1d\x00\x97\x1a\x00\x15\xb0\xaa\x9c'
            b'\xac\xb2@\xa8\xba^\xac\xb2@\x02\xb4\xa2\xa0\xb0\xaa@\xac\xb2`\xb0\xaa\x1b\x04 \x00\x03\x1f\x80'
            b'\x1f\x80\t\x00\x00\x00\x00\x00\x1f\x80\x1f\x80\t\x00\x00\x00\x00\x00\x1f\x80\x1f\x80\t\x00\x00'
            b'\x00\x00\x00\x00\x04\x16\x00\x11\x1f\x80\x1f\x80\t\x00\x00\x00\x00\x00\x1f\x80\x1f\x80\t\x00'
            b'\x00\x00\x00\x00\x00')
        self.assertEqual(f.type.value, 7)
        self.assertEqual(f.first_seq, 2717)
        self.assertEqual(f.seq, 2720)
        self.assertEqual(f.ack, 143)
        self.assertEqual(len(f.pkts), 4)

    def test_encode_decode(self):
        expected = \
            b'COZ\x03RE\x01\x07\x9d\n\xa0\n\x8f\x00\x04\x01\x00\x8f\x04\x1d\x00\x97\x1a\x00\x15\xb0\xaa\x9c' \
            b'\xac\xb2@\xa8\xba^\xac\xb2@\x02\xb4\xa2\xa0\xb0\xaa@\xac\xb2`\xb0\xaa\x1b\x04 \x00\x03\x1f\x80' \
            b'\x1f\x80\t\x00\x00\x00\x00\x00\x1f\x80\x1f\x80\t\x00\x00\x00\x00\x00\x1f\x80\x1f\x80\t\x00\x00' \
            b'\x00\x00\x00\x00\x04\x16\x00\x11\x1f\x80\x1f\x80\t\x00\x00\x00\x00\x00\x1f\x80\x1f\x80\t\x00' \
            b'\x00\x00\x00\x00\x00'
        f = Frame.from_bytes(expected)
        actual = f.to_bytes()
        self.assertEqual(expected, actual)




from typing import List

from .protocol_declaration import FRAME_ID, MIN_FRAME_SIZE, FrameType, PacketType
from .protocol_base import Packet, UnknownPacket, UnknownCommand, UnknownEvent
from .protocol_utils import BinaryReader, BinaryWriter
from .protocol_encoder import Connect, Disconnect, Ping, Unknown0A, ACTION_BY_ID


class Frame(object):

    __slots__ = [
        'type',
        'first_seq',
        'seq',
        'ack',
        'pkts',
    ]

    def __init__(self, type_id: FrameType, first_seq: int, seq: int, ack: int, pkts: List[Packet]) -> None:
        self.type = type_id
        self.first_seq = first_seq
        self.seq = seq
        self.ack = ack
        self.pkts = pkts

    def to_bytes(self) -> bytes:
        writer = BinaryWriter()
        self.to_writer(writer)
        return writer.dumps()

    def _encode_packet(self, pkt, writer) -> None:
        writer.write(pkt.PACKET_ID.value, "B")
        if pkt.PACKET_ID == PacketType.ACTION or pkt.PACKET_ID == PacketType.EVENT:
            writer.write(len(pkt) + 1, "H")
            writer.write(pkt.ID, "B")
        else:
            writer.write(len(pkt), "H")
        writer.write_object(pkt)

    def to_writer(self, writer: BinaryWriter) -> None:
        writer.write_bytes(FRAME_ID)
        writer.write(self.type.value, "B")
        writer.write(self.first_seq, "H")
        writer.write(self.seq, "H")
        writer.write(self.ack, "H")
        if self.type == FrameType.ENGINE or self.type == FrameType.ROBOT:
            for pkt in self.pkts:
                self._encode_packet(pkt, writer)
        elif self.type == FrameType.PING:
            assert len(self.pkts) == 1
            pkt = self.pkts[0]
            assert pkt.PACKET_ID == PacketType.PING
            writer.write_object(pkt)
        elif self.type == FrameType.UNKNOWN_04:
            assert len(self.pkts) == 1
            pkt = self.pkts[0]
            assert pkt.PACKET_ID == PacketType.UNKNOWN
            writer.write_object(pkt)
        elif self.type == FrameType.RESET:
            # No packets
            assert not self.pkts
        elif self.type == FrameType.FIN:
            # No packets
            assert not self.pkts
        else:
            raise NotImplementedError("Unexpected frame type {}.".format(self.type))

    @classmethod
    def from_bytes(cls, buffer: bytes):
        reader = BinaryReader(buffer)
        obj = cls.from_reader(reader)
        return obj

    @classmethod
    def _decode_action(cls, action_id, action_len, reader):
        pkt_class = ACTION_BY_ID.get(action_id)
        if pkt_class:
            res = pkt_class.from_reader(reader)
        else:
            res = UnknownCommand(action_id, reader.read_farray("B", action_len))
        return res

    @classmethod
    def _decode_event(cls, event_id, event_len, reader):
        res = UnknownEvent(event_id, reader.read_farray("B", event_len))
        return res

    @classmethod
    def _decode_packet(cls, pkt_type, pkt_len, reader):
        if pkt_type == PacketType.ACTION:
            action_id = reader.read("B")
            res = cls._decode_action(action_id, pkt_len - 1, reader)
        elif pkt_type == PacketType.EVENT:
            event_id = reader.read("B")
            res = cls._decode_event(event_id, pkt_len - 1, reader)
        elif pkt_type == PacketType.PING:
            res = Ping.from_reader(reader)
        elif pkt_type == PacketType.UNKNOWN_0A:
            res = Unknown0A.from_reader(reader)
        elif pkt_type == PacketType.CONNECT:
            res = Connect.from_reader(reader)
        elif pkt_type == PacketType.DISCONNECT:
            res = Disconnect.from_reader(reader)
        else:
            raise ValueError("Unexpected packet type {}.".format(pkt_type))
        return res

    @classmethod
    def from_reader(cls, reader: BinaryReader):
        if len(reader.buffer) < MIN_FRAME_SIZE:
            raise ValueError("Invalid frame.")

        if reader.buffer[:7] != FRAME_ID:
            raise ValueError("Invalid frame ID.")

        reader.seek_set(7)
        frame_type = FrameType(reader.read("B"))
        first_seq = reader.read("H")
        seq = reader.read("H")
        ack = reader.read("H")
        pkts = []

        if frame_type == FrameType.ENGINE or frame_type == FrameType.ROBOT:
            pkt_seq = first_seq
            while reader.tell() < len(reader):
                pkt_type = PacketType(reader.read("B"))
                pkt_len = reader.read("H")
                pkt = cls._decode_packet(pkt_type, pkt_len, reader)
                pkt.seq = pkt_seq
                pkt.ack = ack
                if not pkt.is_oob():
                    pkt_seq += 1
                pkts.append(pkt)
            assert not seq or seq + 1 == pkt_seq
        elif frame_type == FrameType.PING:
            pkt = Ping.from_reader(reader)
            pkts.append(pkt)
        elif frame_type == FrameType.UNKNOWN_04:
            pkt = UnknownPacket(PacketType.UNKNOWN, reader.buffer[14:])
            pkts.append(pkt)
        elif frame_type == FrameType.RESET:
            # No packets
            assert reader.tell() == len(reader)
        elif frame_type == FrameType.FIN:
            # No packets
            assert reader.tell() == len(reader)
        else:
            raise NotImplementedError("Unexpected frame type {}.".format(frame_type))

        res = cls(frame_type, first_seq, seq, ack, pkts)

        return res



"""

PyCozmo setup script.

"""

import os
import re
import setuptools


def get_package_variable(key):
    fspec = os.path.join("pycozmo", "__init__.py")
    with open(fspec) as f:
        for line in f:
            m = re.match(r"(\S+)\s*=\s*[\"']?(.+?)[\"']?\s*$", line)
            if m and key == m.group(1):
                return m.group(2)
    return None


def get_readme():
    with open("README.md") as f:
        readme = f.read()
    return readme


setuptools.setup(
    name="pycozmo",
    packages=setuptools.find_packages(),
    version=get_package_variable("__version__"),
    license="MIT",
    description="A pure-Python Anki Cozmo communication library.",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    author="Kaloyan Tenchov",
    author_email="zayfod@gmail.com",
    url="https://github.com/zayfod/pycozmo/",
    python_requires=">=3.5",
    install_requires=["dpkt"],
    keywords=["anki", "cozmo", "robot", "robotics"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
    ],
    scripts=[
        "tools/pycozmo_dump.py",
    ],
)




from .frame import Frame
from .client import ROBOT_ADDR, Client
from . robot import *

from . import util
from . import protocol_declaration
from . import protocol_generator
from . import protocol_encoder
from . import protocol_utils


__version__ = "0.1.0"

__all__ = [
    "Frame",
    "ROBOT_ADDR",
    "Client",

    "MIN_HEAD_ANGLE_RAD",
    "MAX_HEAD_ANGLE_RAD",
    "MIN_LIFT_HEIGHT_MM",
    "MAX_LIFT_HEIGHT_MM",
    "LIFT_ARM_LENGTH_MM",
    "LIFT_PIVOT_HEIGHT_MM",
    "MAX_WHEEL_SPEED_MMPS",
]
